from w2v import preprocessing, datasets


def dataset(data, config):
    """Get a dataset.

    Args:
      data: list or reusable iterable.

    Returns:
      w2v.datasets.W2VDataset.
    """
    vocab, freqs, contexts, pairs, _ = preprocessing.preprocess(data, config)
    return datasets.W2VDataset(config, vocab, freqs, contexts, pairs)



from w2v import frequencies, subsampling, tokenization, vocabulary, readers
import collections
import numpy as np


def preprocess(data, config):
    """Pre-process a dataset.

    Args:
      data: reusable iterator or list.
      config: Dictionary.

    Returns:
      vocab: w2v.vocabulary.Vocab.
      freqs: Dictionary.
      contexts: Dictionary.
      pairs: List of tuples.
      stats: Dictionary.
    """
    tokenizer = tokenization.get_tokenizer(config['tokenizer'])
    reader = readers.get_reader(config['reader'], config)

    word_set = set([])
    counts = collections.Counter()
    n = 0

    # first pass
    for line in data:
        tokens = tokenizer(line)
        n += len(tokens)
        word_set.update(tokens)
        counts.update(tokens)

    vocab = vocabulary.Vocab(config['name'], word_set)
    freqs = frequencies.freqs(counts, n)
    for token in vocabulary.extra_tokens():
        freqs[token] = 0.
    drop_probs = subsampling.p(freqs, config['t'])
    subsampler = subsampling.SubSampler(drop_probs)
    contexts = {}
    pairs = []

    # second pass
    for line in data:
        tokens = tokenizer(line)
        tokens = subsampler(tokens)
        new_contexts, new_pairs = reader(tokens)
        # update contexts
        for token, context in new_contexts.items():
            if token not in contexts.keys():
                contexts[token] = readers.Context(token)
            contexts[token].update(context)
        pairs += new_pairs

    # gather and report stats
    context_lens = [len(c) for c in contexts.values()]
    avg_context_len = np.average(context_lens)
    stats = {
        'corpus_len': n,
        'vocab_size': len(vocab),
        'training_pairs': len(pairs),
        'avg_context_len': avg_context_len}
    print('Pre-processing complete.')
    print('Corpus length: %s' % n)
    print('Vocab size: %s' % len(vocab))
    print('Training pairs: %s' % len(pairs))
    print('Average context size: %s' % avg_context_len)

    return vocab, freqs, contexts, pairs, stats



"""For creating and managing vocabularies."""


# Defaults for
SOS = '<S>'
EOS = '</S>'
PAD = "<PAD>"
UNK = "<UNK>"
# LBR and RBR are for handling S-Expressions, though usage is fairly obscure
LBR = '('
RBR = ')'


def extra_tokens():
    """Get a list of extra tokens added to Vocab.

    Returns:
      List of strings.
    """
    return [SOS, EOS, PAD, UNK, LBR, RBR]


class Vocab:
    """Wrapper for a vocab dict.

    Exposes indexing for both strings and integers for forward and reverse
    lookup.

    Examples:
      word_id = vocab['word']
      word = vocab[word_id]

    Attributes:
      name: String.
      ix: Dictionary, mapping word keys to index values {token: id}.
      word: Dictionary, mapping index keys to word values {id: token}.
      n: Integer, the number of words in the vocabulary.
      pad: String, the padding token. Default is "<PAD>".
      unk: String, the unknown token.
      sos: String, the start of sequence token. Default is "<S>".
      eos: String, the end of sequence token. Default is "</S>".
      pad_ix: Integer, the index of the pad token.
      unk_ix: Integer, the index of the unknown token.
      sos_ix: Integer, the index of the sos token.
      eos_ix: Integer, the index of the eos token.
    """

    def __init__(self, name, words, pad=PAD, unk=UNK, sos=SOS, eos=EOS):
        """Create a new Vocab.

        Args:
          name: String, the name of the vocabulary, e.g. the dataset name.
          words: Set of strings.
          pad: String, the padding token. Default is global PAD.
          unk: String, the unknown token. Default is global UNK.
          sos: String, the start of sequence token. Default is global SOS.
          eos: String, the end of sequence token. Default is global EOS.
        """
        self.name = name
        words.update([sos, eos, pad, unk, LBR, RBR])
        self.ix = dict(zip(words, range(1, len(words) + 1)))
        # Make sure PAD is ix 0
        self.ix[pad] = 0
        self.word = {v: k for k, v in self.ix.items()}
        self.n = len(self.ix)
        self.pad = pad
        self.unk = unk
        self.sos = sos
        self.eos = eos
        self.pad_ix = self[pad]
        self.unk_ix = self[unk]
        self.sos_ix = self[sos]
        self.eos_ix = self[eos]

    def __getitem__(self, item):
        # Item may be either an int or a string (forward or reverse lookup)
        if isinstance(item, int):
            return self.word[item]
        elif isinstance(item, str):
            return self.ix[item]
        else:
            raise ValueError('Unexpected type: %s.' % type(item))

    def __len__(self):
        return self.n

    def __repr__(self):
        return 'Vocab for %s with %s words.' % (self.name, self.n)



"""Data structures for inputs."""


class IterableAdapter:
    """Wrapper for reusing iterables.

    https://stackoverflow.com/questions/1271320/
    resetting-generator-object-in-python

    michaelsnowden's answer.
    """

    def __init__(self, iterator_factory):
        self.iterator_factory = iterator_factory

    def __iter__(self):
        return self.iterator_factory()



"""Tokenization functions."""
import nltk
import spacy


def get_tokenizer(name):
    if name == 'nltk':
        return NLTKTokenizer()
    elif name == 'spacy':
        return SpacyTokenizer()
    else:
        raise ValueError('Unexpected tokenizer name "%r"' % name)


class Tokenizer:
    """Tokenizer interface declaration."""

    def __call__(self, text):
        return self.tokenize(text)

    def tokenize(self, text):
        raise NotImplementedError


class NLTKTokenizer(Tokenizer):
    """NLTK tokenizer."""

    def __init__(self):
        super(NLTKTokenizer, self).__init__()

    def tokenize(self, text):
        return nltk.word_tokenize(text)


class SpacyTokenizer(Tokenizer):
    """Spacy tokenizer."""

    def __init__(self):
        super(SpacyTokenizer, self).__init__()
        self.nlp = spacy.load('en')

    def tokenize(self, text):
        doc = self.nlp(text)
        return [t.text for t in doc]



"""Subsampling frequent words."""
import numpy as np


def p(freqs, t=0.0001):
    """Get probabilities for sampling a word based on frequencies.

    Args:
      freqs: collections.Counter.
      t: float, constant for the function.

    Returns:
      Dictionary.
    """
    keys = freqs.keys()
    probs = np.maximum(0., 1. - np.sqrt(t / np.array(list(freqs.values()))))
    return dict(zip(keys, probs))


class SubSampler:

    def __init__(self, probs):
        self.probs = probs

    def __call__(self, items):
        return [i for i in items
                if not np.random.binomial(n=1, p=self.probs[i])]



"""Calculating word frequencies in corpora."""
import numpy as np


def freqs(counts, n):
    keys = counts.keys()
    tok_counts = np.array(list(counts.values()))
    freqs = tok_counts / n
    return dict(zip(keys, freqs))



"""Negative sampling."""
import numpy as np


class NegativeSampler:

    def __init__(self, frequencies, contexts, num_negs):
        """Create a new NegativeSampler.

        Args:
          frequencies: List of integers, the frequencies of each word,
            sorted in word index order.
          contexts: Dictionary.
          num_negs: Integer, how many to negatives to sample.
        """
        self.n = len(frequencies)
        self.values = list(frequencies.keys())
        self.contexts = contexts
        self.num_negs = num_negs
        self.distribution = self.p(list(frequencies.values()))

    def __call__(self, input):
        """Get negative samples.

        Args:
          input: object, e.g. a string or an integer ix.

        Returns:
          List of objects not in the context.
        """
        context = self.contexts[input]
        max_negs = self.n - len(context) - 1
        k = min(self.num_negs, max_negs)
        neg_ixs = np.random.choice(
            self.n,
            size=k,
            p=self.distribution,
            replace=False)
        negs = [self.values[ix] for ix in neg_ixs]
        # make sure we haven't sampled center word or its context
        invalid = [input] + list(context)
        for i, neg in enumerate(negs):
            while negs[i] in invalid:
                new_ix = np.random.choice(self.n,
                                          size=1,
                                          p=self.distribution)[0]
                new_neg = self.values[new_ix]
                if new_neg not in negs:
                    negs[i] = new_neg
        return negs

    def p(self, freqs):
        """Determine the probability distribution for negative sampling.

        Args:
          freqs: List of integers.

        Returns:
          numpy.ndarray.
        """
        freqs = np.array(freqs)
        return np.power(freqs, 3/4) / np.sum(np.power(freqs, 3/4))



"""Reading contexts from data."""


def get_reader(name, config):
    if name == 'window':
        return WindowReader(config['k'])
    else:
        raise ValueError('Unexpected reader name "%r"' % name)


class Context:
    """Interface declaration and base implementation.

    Attributes:
      x: object, the center object (e.g. token).
      context: set.
    """

    def __init__(self, x):
        self.x = x
        self.context = set([])

    def __iter__(self):
        for x in self.context:
            yield x

    def __len__(self):
        return len(self.context)

    def __repr__(self):
        return 'x: %s\ncontext: %s' % (self.x, ' '.join(self.context))

    def update(self, context):
        """Update this context.

        Args:
          context: List of objects in the context.
        """
        context = [c for c in context if c != self.x]
        self.context.update(context)


class Reader:
    """Interface declaration."""

    def __call__(self, inputs, *args):
        return self.contexts_and_pairs(inputs)

    def contexts_and_pairs(self, inputs):
        raise NotImplementedError


class WindowReader(Reader):
    """Read a context window around tokens in text."""

    def __init__(self, k):
        """Create a new WindowReader.

        Args:
          k: Integer, window size.
        """
        self.k = k

    def contexts_and_pairs(self, tokens):
        contexts = {}
        pairs = []
        for i, token in enumerate(tokens):
            context = self.token_context(i, tokens)
            if token not in contexts.keys():
                contexts[token] = Context(token)
            contexts[token].update(context)
            pairs += list(zip([token] * len(context), context))
        return contexts, pairs

    def token_context(self, center_ix, tokens):
        """Get left and right contexts for the token.

        Args:
          center_ix: Integer.
          tokens: List of strings.

        Returns:
          contexts: Dictionary.
        """
        left_start = max(0, center_ix - self.k)
        left_end = center_ix
        left_context = tokens[left_start:left_end]
        right_start = center_ix + 1
        right_end = min(len(tokens), center_ix + 1 + self.k)
        right_context = tokens[right_start:right_end]
        return left_context + right_context



"""Wrappers for word2vec data."""
from torch.utils.data import dataset


class W2VDataset(dataset.Dataset):
    """Base W2V dataset.

    Attributes:
      dict: w2v.vocabulary.Vocab.
      freqs: Dictionary.
      contexts: Dictionary.
      pairs: List of tuples.
    """

    def __init__(self, config, vocab, freqs, contexts, pairs):
        """Create a new W2VDataset.

        Args:
          config: Dictionary.
          vocab: w2v.vocabulary.Vocab.
          freqs: Dictionary.
          contexts: Dictionary.
          pairs: List of tuples.
        """
        for key, value in config.items():
            setattr(self, key, value)
        self.vocab = vocab
        self.freqs = freqs
        self.contexts = contexts
        self.pairs = pairs
        self._length = len(pairs)

    def __getitem__(self, i):
        return self.pairs[i]

    def __len__(self):
        return self._length



"""Collate functions."""


class Collate:
    """Callable collate function."""

    def __init__(self, neg_sampler):
        """Create a new Collate.

        Args:
          neg_sampler: w2v.neg_sampling.NegativeSampler.
        """
        self.sampler = neg_sampler

    def __call__(self, pairs):
        """Call the collate function.

        Args:
          pairs: List of tuples of strings (center, context).

        Returns:
          centers: List of strings.
          contexts: List of strings.
          negs: List of lists of strings.
        """
        batch_size = len(pairs)
        centers = [x[0] for x in pairs]
        contexts = [x[1] for x in pairs]
        negs = []
        for i in range(batch_size):
            negs.append(self.sampler(centers[i]))
        # a generic return type for base class (can override with specifics)
        return centers, contexts, negs
